# coding=utf-8

import random

import torch
from transformers import MarianTokenizer, MarianMTModel
from googletrans import Translator

from promptcraft import styleformer
from promptcraft import parrot


class SentencePerturb:
    """
    Sentence-level Prompt Perturbation
    SentencePerturb class for directly manipulating a sentence
    Style transformation credits: https://github.com/PrithivirajDamodaran/Styleformer
    Paraphrasing credits: https://github.com/PrithivirajDamodaran/Parrot_Paraphraser
    """

    def __init__(self, sentence):
        # The sentence to be perturbed
        self.sentence = sentence

    def formal(self):
        # Input sentence
        sentence = self.sentence[:]

        # Casual to Formal
        style = styleformer.StyleFormer(style=0)

        try:
            # Implement the style transformation
            sentence = style.transfer(sentence)
        except:
            print("Error in formal")
            return None

        return sentence

    def casual(self):
        # Input sentence
        sentence = self.sentence[:]

        # Formal to Casual
        style = styleformer.StyleFormer(style=1)

        try:
            # Implement the style transformation
            sentence = style.transfer(sentence)
        except:
            print("Error in casual")
            return None

        return sentence

    def passive(self):
        # Input sentence
        sentence = self.sentence[:]

        # Active to Passive
        style = styleformer.StyleFormer(style=2)

        try:
            # Implement the style transformation
            sentence = style.transfer(sentence)
        except:
            print("Error in passive")
            return None

        return sentence

    def active(self):
        # Input sentence
        sentence = self.sentence[:]

        # Passive to Active
        style = styleformer.StyleFormer(style=3)

        try:
            # Implement the style transformation
            sentence = style.transfer(sentence)
        except:
            print("Error in active")
            return None

        return sentence

    def back_translation_hugging_face(self):
        # Input sentence
        sentence = self.sentence[:]

        # Load the pre-trained English to German and German to English translation models
        en_de_name = "Helsinki-NLP/opus-mt-en-de"
        de_en_name = "Helsinki-NLP/opus-mt-de-en"

        tokenizer_en_de = MarianTokenizer.from_pretrained(en_de_name)
        model_en_de = MarianMTModel.from_pretrained(en_de_name, torch_dtype=torch.float32)

        tokenizer_de_en = MarianTokenizer.from_pretrained(de_en_name)
        model_de_en = MarianMTModel.from_pretrained(de_en_name, torch_dtype=torch.float32)

        try:
            # Translate from English to German
            trans_german = model_en_de.generate(**tokenizer_en_de(sentence,
                                                                  return_tensors="pt",
                                                                  padding=True,
                                                                  truncation=True))
            bt_german = tokenizer_en_de.decode(trans_german[0], skip_special_tokens=True)
        except:
            print("Error in back_translation_hugging_face")
            return None

        try:
            # Translate from German back to English
            trans_english = model_de_en.generate(**tokenizer_de_en(bt_german,
                                                                   return_tensors="pt",
                                                                   padding=True,
                                                                   truncation=True))
            bt_english = tokenizer_de_en.decode(trans_english[0], skip_special_tokens=True)
        except:
            print("Error in back_translation_hugging_face")
            return None

        return bt_english

    def back_translation_google(self):
        # Input sentence
        sentence = self.sentence[:]

        # Create a translator object
        translator = Translator()

        try:
            # Translate the sentence to German
            german = translator.translate(sentence, dest='de').text
        except:
            print("Error in back_translation_google")
            return None

        try:
            # Translate the German translation back to English
            back_english = translator.translate(german, dest='en').text
        except:
            print("Error in back_translation_google")
            return None

        return back_english

    def paraphrase(self):
        # Input sentence
        sentence = self.sentence[:]

        # Load Paraphraser
        paraphraser = parrot.Parrot()

        try:
            # Get paraphrased sentences
            paraphrases = paraphraser.augment(input_phrase=sentence)
            sentence = random.sample(paraphrases, 1)[0][0]

            return sentence
        except:
            return None



# coding=utf-8

import re
import random

import nltk
from nltk.corpus import stopwords
from nltk.corpus import wordnet

nltk.download('stopwords')
nltk.download('punkt')
nltk.download('wordnet')

special_symbols = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"

# Stop words list
stop_words = ['i', 'me', 'my', 'myself', 'we', 'our',
              'ours', 'ourselves', 'you', 'your', 'yours',
              'yourself', 'yourselves', 'he', 'him', 'his',
              'himself', 'she', 'her', 'hers', 'herself',
              'it', 'its', 'itself', 'they', 'them', 'their',
              'theirs', 'themselves', 'what', 'which', 'who',
              'whom', 'this', 'that', 'these', 'those', 'am',
              'is', 'are', 'was', 'were', 'be', 'been', 'being',
              'have', 'has', 'had', 'having', 'do', 'does', 'did',
              'doing', 'a', 'an', 'the', 'and', 'but', 'if', 'or',
              'because', 'as', 'until', 'while', 'of', 'at',
              'by', 'for', 'with', 'about', 'against', 'between',
              'into', 'through', 'during', 'before', 'after',
              'above', 'below', 'to', 'from', 'up', 'down', 'in',
              'out', 'on', 'off', 'over', 'under', 'again',
              'further', 'then', 'once', 'here', 'there', 'when',
              'where', 'why', 'how', 'all', 'any', 'both', 'each',
              'few', 'more', 'most', 'other', 'some', 'such', 'no',
              'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too',
              'very', 's', 't', 'can', 'will', 'just', 'don',
              'should', 'now', '']


def get_synonyms(word):
    """
    Get synonyms for a word
    """
    synonyms = []
    for syn in wordnet.synsets(word):
        for lemma in syn.lemmas():
            synonyms.append(lemma.name())

    return synonyms


class WordPerturb:
    """
    Word-level Prompt Perturbation
    WordPerturb class for manipulating words in a sentence

    NOTE: the number of words in a sentence is only the valid words
    without considering spaces, special symbols, and punctuations
    """

    def __init__(self, sentence, level):
        # Original sentence
        self.sentence = sentence

        # Tokenize the sentence into words
        # self.words = word_tokenize(sentence)
        self.words = re.findall(r"\w+|\s+|[^\w\s]", self.sentence, re.UNICODE)

        # The perturbation level to be implemented
        # use the regular expression pattern \b\w+\b to match words
        # \b matches word boundaries
        # \w+ matches one or more word characters
        self.valid_words = re.findall(r'\b\w+\b', self.sentence)
        self.num = int(len(self.valid_words) * level)

        # Get the set of English stop words
        self.stop_words = set(stopwords.words('english'))
        self.stop_words.add(" ")
        for symbol in special_symbols:
            self.stop_words.add(symbol)

        for stop in stop_words:
            self.stop_words.add(stop)

    def synonym_replacement(self):
        """
        Randomly choose n words from the sentence that are not stop words.
        Replace each of these words with one of its synonyms chosen at random.

        # Problem 1: Without any synonyms
        # Problem 2: Fewer positions than needed positions
        """
        # Convert the sentence to a list of words for easier manipulation
        sen_list = self.words[:]

        # Create a list of positions that correspond to the non-stop words
        positions = [i for i, word in enumerate(sen_list) if word.lower() not in self.stop_words]

        # Remove those positions that don't have any synonym
        for i in range(len(positions)):
            word_ori = self.words[positions[i]]
            synonyms = get_synonyms(word_ori)
            if len(synonyms) == 0:
                positions[i] = -1

        while -1 in positions:
            positions.remove(-1)

        # Randomly sample `self.num` positions
        if self.num >= len(positions):
            pass
        else:
            positions = random.sample(positions, self.num)

        # Replace chosen words with random synonyms
        for index in positions:
            word_ori = self.words[index]
            synonyms = get_synonyms(word_ori)

            # Randomly retrieve one synonym from all the synonyms
            synonym = random.choice(synonyms)
            if '_' in synonym:
                synonym = synonym.replace('_', ' ')

            sen_list[index] = synonym

        # Join the modified words back into a string
        sen_list = ''.join(sen_list)

        return sen_list

    def word_insertion(self):
        """
        Find a random synonym of a random word in the sentence that is not a stop word.
        Insert that synonym into a random position in the sentence.
        Do this n times.
        """
        # Convert the sentence to a list of words for easier manipulation
        sen_list = self.words.copy()

        # Create a list of positions that correspond to the non-stop words
        positions = [i for i, word in enumerate(sen_list) if word.lower() not in self.stop_words]
        non_sw_posi = positions.copy()

        # Remove those positions that don't have any synonym
        for i in range(len(positions)):
            word_ori = self.words[positions[i]]
            synonyms = get_synonyms(word_ori)
            if len(synonyms) == 0:
                positions[i] = -1

        while -1 in positions:
            positions.remove(-1)

        # Randomly sample `self.num` positions
        if self.num >= len(positions):
            pass
        else:
            positions = random.sample(positions, self.num)

        # Initialize an empty list to store the modified sentence.
        sen_init = []
        positions_insert = random.sample(non_sw_posi, self.num)
        for index, word in enumerate(sen_list):
            # Check if the current word's index is in the list of chosen indices.
            if index in positions_insert:
                # Insert a random character in front of the chosen character.
                random_index = random.sample(positions, 1)
                word_ori = self.words[random_index[0]]
                synonyms = get_synonyms(word_ori)

                # Randomly retrieve one synonym from all the synonyms
                word_insert = random.choice(synonyms)
                if '_' in word_insert:
                    word_insert = word_insert.replace('_', ' ')

                sen_init.append(word_insert)
                sen_init.append(" ")

            sen_init.append(word)

        # Convert the modified list back to a string.
        sen_init = ''.join(sen_init)

        return sen_init

    def word_swap(self):
        """
        Randomly choose two words in the sentence and swap their positions.
        Do this n times.
        """
        # Convert the sentence to a list of words for easier manipulation
        sen_list = self.words.copy()

        # Create a list of positions that correspond to non-stop words
        positions = [i for i, word in enumerate(sen_list) if word.lower() not in self.stop_words]

        # Choose random positions 1 and 2 for swapping
        position_1 = random.sample(positions, self.num)
        position_2 = random.sample(positions, self.num)

        # Perform the character swaps
        for i in range(self.num):
            sen_list[position_1[i]], sen_list[position_2[i]] = sen_list[position_2[i]], sen_list[position_1[i]]

        # Convert the list of characters back to a string
        sen_list = ''.join(sen_list)

        return sen_list

    def word_deletion(self):
        """
        Each word in the sentence can be randomly removed with probability p.
        """
        # Convert the sentence to a list of words for easier manipulation
        sen_list = self.words.copy()

        # Create a list of positions that correspond to non-stop words
        positions = [i for i, word in enumerate(sen_list) if word.lower() not in self.stop_words]

        # Randomly sample `self.num` positions from all positions
        positions = random.sample(positions, self.num)

        # Sort the indices in reverse order so that deletion doesn't affect subsequent indices
        positions.sort(reverse=True)

        # Delete the selected characters
        for index in positions:
            del sen_list[index]

        # Join the modified characters back into a string
        sen_list = ''.join(sen_list)

        return sen_list

    def insert_punctuation(self):
        """
        Randomly insert punctuation in the sentence with probability p.
        """
        # Convert the sentence to a list of words for easier manipulation
        sen_list = self.words.copy()

        # Create a list of positions that correspond to non-stop words
        positions = [i for i, word in enumerate(sen_list) if word.lower() not in self.stop_words]

        # Randomly sample `self.num` positions from all positions
        positions = random.sample(positions, self.num)

        # Initialize an empty list to store the modified sentence.
        sen_init = []
        for index, word in enumerate(self.words):
            sen_init.append(word)
            # Check if the current character's index is in the list of chosen indices.
            if index in positions:
                # Insert a random character in front of the chosen character.
                word_insert = random.choice(special_symbols)
                sen_init.append(word_insert)

        # Convert the modified list back to a string.
        sen_init = ''.join(sen_init)

        return sen_init

    def word_split(self):
        """
        Randomly split a word to two tokens randomly
        """
        # Convert the sentence to a list of words for easier manipulation
        sen_list = self.words.copy()

        # Create a list of positions that correspond to non-stop words
        positions = [i for i, word in enumerate(sen_list) if word.lower() not in self.stop_words]

        # Randomly sample `self.num` positions from all positions
        positions = random.sample(positions, self.num)

        # Initialize an empty list to store the modified sentence
        sen_init = []
        for index, word in enumerate(self.words):
            if index in positions:
                if len(word) > 1:
                    # Split the word into two pieces
                    indice = random.randint(1, len(word) - 1)

                    # Get the left and right subwords
                    left_subword = word[:indice]
                    right_subword = word[indice:]

                    # Append the left and right subwords
                    sen_init.append(left_subword)
                    sen_init.append(" ")
                    sen_init.append(right_subword)
                else:
                    pass
            else:
                sen_init.append(word)

        # Convert the modified list back to a string.
        sen_init = ''.join(sen_init)

        return sen_init



# coding=utf-8

import random


# Define common characters
char_common = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~ "


# Define the USA keyboard layout mapping for nearby characters
keyboard_layout = {
    'a': 'qwsxz',
    'b': 'vghn',
    'c': 'xdfv',
    'd': 'serfcx',
    'e': 'wrdasf34',
    'f': 'drtgvc',
    'g': 'ertyhvf',
    'h': 'rtyujbg',
    'i': 'ujko89',
    'j': 'tyuiknh',
    'k': 'yuiojl',
    'l': 'uiojkm',
    'm': 'njk',
    'n': 'bhjm',
    'o': 'iklp90',
    'p': 'ol0',
    'q': 'aw12',
    'r': 'etdf45',
    's': 'awedxz',
    't': 'ryfg56',
    'u': 'yihj78',
    'v': 'cfgb',
    'w': 'qase23',
    'x': 'zasdc',
    'y': 'tghu67',
    'z': 'asx',
    '0': '9op',
    '1': '2q',
    '2': '13qw',
    '3': '24we',
    '4': '35er',
    '5': '46rt',
    '6': '57ty',
    '7': '68yu',
    '8': '79ui',
    '9': '80io'
}


ocr_errors = {
    'a': ['A', '4', '@', '∀', '^', 'Λ', '∆', 'λ', 'a', 'o', 'e'],
    'b': ['B', '8', '6', 'b', 'd', '|3', 'ß', 'Β', 'ƃ'],
    'c': ['C', 'c', 'e', 'o', '0', '©', '<', '[', '(', '¢'],
    'd': ['D', 'd', 'o', '0', '∂', 'δ', 'o|', 'cl', 'þ', '])', 'ð'],
    'e': ['E', '3', 'e', '∑', '€', '∈', '∃', '£', 'έ', 'Σ', 'ë'],
    'f': ['F', 'f', 't', '7', '+', 'ϝ', 'ʃ', 'ƒ', 'λ', 'ƭ', 'ρ'],
    'g': ['G', 'g', '9', '6', 'q', '&', 'gq', 'ʛ', 'ğ', 'ǥ', 'φ'],
    'h': ['H', 'h', '4', '5', '#', 'η', 'н', 'ħ', 'ђ', 'λ', 'н'],
    'i': ['I', '1', 'l', '|', '!', '∫', 'ί', 'Ǐ', 'ï', 'Σ', 'н'],
    'j': ['J', 'j', '1', '7', 'ij', 'ϳ', 'j', 'ʝ', 'ǰ', 'ן', 'ɉ'],
    'k': ['K', 'k', 'l', '|<', '|{', 'κ', 'ќ', 'κ', 'ƙ', 'ķ', 'л'],
    'l': ['L', 'l', '1', 'i', '|', '!', 'ℓ', '£', '£', 'ί', 'լ'],
    'm': ['M', 'm', 'n', 'rn', 'w', 'vv', '^^', 'mm', 'μ', 'λ', 'м'],
    'n': ['N', 'n', 'm', 'r', 'π', 'h', 'ח', 'й', 'ŋ', 'π', 'η'],
    'o': ['O', '0', 'o', 'θ', 'Ø', 'σ', 'φ', 'Φ', 'ο', '∞', 'ם'],
    'p': ['P', 'p', '|o', '|D', 'ρ', 'ρ', 'þ', 'φ', 'ρ', '¶', 'р'],
    'q': ['Q', 'q', 'g', '9', 'ℚ', 'ѳ', 'ϙ', 'ɋ', 'ք', 'Ɋ', 'ͼ'],
    'r': ['R', 'r', 'l', '1', '|', '7', 'я', 'г', '®', 'λ', 'г'],
    's': ['S', '5', 's', '§', 'z', '$', '∫', 'ʂ', 'ѕ', 'š', 'ʃ'],
    't': ['T', 't', '+', '7', '†', 'τ', 'т', 'ţ', 'ť', '±', 'π'],
    'u': ['U', 'u', 'v', 'w', 'µ', 'υ', 'ц', 'щ', 'μ', 'µ', 'บ'],
    'v': ['V', 'v', 'w', 'u', '∨', 'ν', '∀', 'ѵ', 'ν', 'в', 'ט'],
    'w': ['W', 'w', 'vv', 'uu', 'wv', 'wu', 'ш', 'ω', 'ώ', 'ω', 'ш'],
    'x': ['X', 'x', '+', '*', '×', 'χ', 'ж', 'ж', 'א', 'ж', 'ж'],
    'y': ['Y', 'y', 'γ', 'у', '¥', 'ү', 'ч', 'у', 'λ', 'ф', 'י'],
    'z': ['Z', 'z', '2', '7', 'ʒ', 'я', 'ʒ', 'ж', 'ʑ', 'ʐ', 'ζ'],
    '1': ['l', 'i', '!', '|'],
    '2': ['5', 'z', 's', '3'],
    '3': ['8', '2', '9'],
    '4': ['h', 'k', 'b', 'l', 'r'],
    '5': ['s', 'z', '2', '3', '8'],
    '6': ['b', '8', 'g', '9', '0'],
    '7': ['t', 'f', '1', 'h', 'k', 'b', 'l', 'r'],
    '8': ['5', '2', '3', '9'],
    '9': ['q', 'g', 'p', '6', '0', '8'],
    '0': ['o', 'd', 'c', 'p', 'g', 'q'],
    'A': ['4', '6', '8', '9', '0'],
    'B': ['6', '8'],
    'C': ['9', '0', 'e', 'o'],
    'D': ['0', '9', 'a', 'o'],
    'E': ['3', '8'],
    'F': ['5', '7', 't'],
    'G': ['6', '9', 'q', 'y'],
    'H': ['4', '6', '8'],
    'I': ['1', 'l', '!', '|'],
    'J': ['1', '7'],
    'K': ['1', '4'],
    'L': ['1', 'i', '!', '|'],
    'M': ['n'],
    'N': ['m'],
    'O': ['0', 'c', 'd', 'q'],
    'P': ['9', 'q', 'b'],
    'Q': ['9', '0', 'o', 'p', 'g'],
    'R': ['7', '5', 't'],
    'S': ['5', '9', '2'],
    'T': ['7', '1', 'f'],
    'U': ['v'],
    'V': ['u'],
    'W': ['vv'],
    'X': ['4', '8'],
    'Y': ['g', '9'],
    'Z': ['2', '5']
    # '!': ['1', 'i', 'l', '|', '7'],
    # '|': ['l', 'i', '1', '!'],
    # '/': ['1', 'l', 'i', '!', '|'],
    # '(': ['8', '6', '9', 't', 'f', '7'],
    # '[': ['c', 'e', 'o', '(', '¢', '©'],
    # '¢': ['c', 'e', 'o', 'o', 'a', 'c', 'c', 'o', 'd', 'e', 'g', 's', 'q'],
    # '©': ['c', 'e', 'o', 'o', 'a', 'c', 'c', 'o', 'd', 'e', 'g', 's', 'q'],
    # '¥': ['7', '1', 'f', 'h', 'k', 'b', 'l', 't'],
    # '€': ['c', 'e', 'o'],
    # '£': ['6', '8'],
    # '$': ['5', '8', '2'],
}


class CharacterPerturb:
    """
    Character-level Prompt Perturbation
    CharacterPerturb class for manipulating character in a sentence
    """

    def __init__(self, sentence, level):
        # The sentence to be perturbed
        self.data = sentence
        # The perturbation level to be implemented
        self.percent = level

    def character_replacement(self):
        """
        Randomly replace `self.percent` percentage characters from the self.data
        """
        # Convert the sentence to a list of characters for easier manipulation
        char_list = list(self.data)

        # Calculate the number of characters to choose
        num = int(len(char_list) * self.percent)

        # Choose random positions to substitute
        positions = random.sample(range(len(self.data)), num)

        # Substitute the chosen characters with random characters
        for index in positions:
            char_re = random.choice(char_common)
            char_list[index] = char_re

        # Join the modified characters back into a string
        sentence = ''.join(char_list)

        return sentence

    def character_deletion(self):
        """
        Randomly delete `self.percent` percentage characters from the self.data
        """
        # Convert the sentence to a list of characters for easier manipulation
        char_list = list(self.data)

        # Calculate the number of characters to choose
        num = int(len(char_list) * self.percent)

        # Choose random positions to delete
        positions = random.sample(range(len(self.data)), num)

        # Sort the indices in reverse order so that deletion doesn't affect subsequent indices
        positions.sort(reverse=True)

        # Delete the selected characters
        for index in positions:
            char_list[index] = "X"
            # char_list[index] = "[X]"
            # del char_list[index]

        # Join the modified characters back into a string
        sentence = ''.join(char_list)

        return sentence

    def character_insertion(self):
        """
        Randomly insert `self.percent` percentage characters to the self.data
        """
        # Convert the sentence to a list of characters for easier manipulation
        char_list = list(self.data)

        # Calculate the number of characters to choose
        num = int(len(char_list) * self.percent)

        # Choose random positions to insert
        positions = random.sample(range(len(self.data)), num)

        # Initialize an empty list to store the modified sentence.
        sentence = []
        for index, char in enumerate(self.data):
            # Check if the current character's index is in the list of chosen indices.
            if index in positions:
                # Insert a random character in front of the chosen character.
                char_insert = random.choice(char_common)
                sentence.append(char_insert)

            sentence.append(char)

        # Convert the modified list back to a string.
        sentence = ''.join(sentence)

        return sentence

    def character_swap(self):
        """
        Randomly swap `self.percent` percentage characters in the self.data
        NOTE: including self-swapping
        """
        # Convert the sentence to a list of characters for easier manipulation
        char_list = list(self.data)

        # Calculate the number of characters to choose
        num = int(len(char_list) * self.percent)

        # Choose random positions 1 and 2 for swapping
        position_1 = random.sample(range(len(self.data)), num)
        position_2 = random.sample(range(len(self.data)), num)

        # Perform the character swaps
        for i in range(num):
            char_list[position_1[i]], char_list[position_2[i]] = char_list[position_2[i]], char_list[position_1[i]]

        # Convert the list of characters back to a string
        sentence = ''.join(char_list)

        return sentence

    def keyboard_typos(self):
        """
        Randomly substitute `self.percent` percentage characters in the self.data
        with a randomly chosen character which is near the original character in the Keyboard (USA Full-size Layout)
        NOTE:
            (1) We applied `keyboard_distance=1`, i.e., the nearest character, number, or samples.
            (2) If it is a character, we randomly chose lowercase or uppercase.
        """
        # Convert the sentence to a list of characters for easier manipulation
        char_list = list(self.data)

        # Calculate the number of characters to choose
        num = int(len(char_list) * self.percent)

        # Create a list of positions that correspond to characters or numbers
        positions = [i for i, char in enumerate(self.data) if char.isalnum()]
        positions = random.sample(positions, num)

        # Substitute the chosen characters with near keyboard characters
        for index in positions:
            char_ori = char_list[index]
            if char_ori.lower() in keyboard_layout:
                char_re = random.choice(keyboard_layout[char_ori.lower()])
                char_re = random.choice([char_re.lower(), char_re.upper()])
                char_list[index] = char_re

        # Join the modified characters back into a string
        sentence = ''.join(char_list)

        return sentence

    def optical_character_recognition(self):
        """
        Randomly substitute `self.percent` percentage characters in the self.data
        with a common OCR map error
        """
        # Convert the sentence to a list of characters for easier manipulation
        char_list = list(self.data)

        # Calculate the number of characters to choose
        num = int(len(char_list) * self.percent)

        # Create a list of positions that correspond to characters or numbers
        positions = [i for i, char in enumerate(self.data) if char.isalnum()]
        positions = random.sample(positions, num)

        # Substitute the chosen characters with common OCR errors
        for index in positions:
            char_ori = char_list[index]
            char_re = random.choice(ocr_errors[char_ori])
            char_list[index] = char_re

        # Join the modified characters back into a string
        sentence = ''.join(char_list)

        return sentence



# coding=utf-8
#
# Copyright 2021: https://github.com/PrithivirajDamodaran/Styleformer
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modified by Shuyue Jia

import torch
from transformers import (
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
)

adequacy_tag = 'prithivida/parrot_adequacy_model'
ftc_tag = "prithivida/formal_to_informal_styletransfer"
ctf_tag = "prithivida/informal_to_formal_styletransfer"
atp_tag = "prithivida/active_to_passive_styletransfer"
pta_tag = "prithivida/passive_to_active_styletransfer"


class Adequacy:

    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(adequacy_tag)
        self.model = AutoModelForSequenceClassification.from_pretrained(adequacy_tag, torch_dtype=torch.float32)

    def filter(self, input_phrase, paraphrases, threshold, device="cpu"):
        top_phrases = []
        for phrase in paraphrases:
            x = self.tokenizer(input_phrase, phrase, return_tensors='pt', max_length=128, truncation=True)
            self.model = self.model.to(device)

            logits = self.model(**x).logits
            probs = logits.softmax(dim=1)
            prob_label = probs[:, 1]

            score = prob_label.item()
            if score >= threshold:
                top_phrases.append(phrase)

        return top_phrases

    def score(self, input_phrase, paraphrases, threshold, device="cpu"):
        scores = {}
        for phrase in paraphrases:
            x = self.tokenizer(input_phrase, phrase, return_tensors='pt', max_length=128, truncation=True)
            x = x.to(device)
            self.model = self.model.to(device)

            logits = self.model(**x).logits
            probs = logits.softmax(dim=1)
            prob_label = probs[:, 1]

            score = prob_label.item()
            if score >= threshold:
                scores[phrase] = score

        return scores


class StyleFormer:

    def __init__(self, style=0):
        self.style = style

    def transfer(self, input_phrase, filter=0.95, candidates=5):
        device = "cpu"

        if self.style == 0:
            output = self._casual_to_formal(input_phrase, device, filter, candidates)
            return output

        elif self.style == 1:
            output = self._formal_to_casual(input_phrase, device, filter, candidates)
            return output

        elif self.style == 2:
            output = self._active_to_passive(input_phrase, device)
            return output

        elif self.style == 3:
            output = self._passive_to_active(input_phrase, device)
            return output

    def _formal_to_casual(self, input_phrase, device, filter, candidates):
        self.ftc_tokenizer = AutoTokenizer.from_pretrained(ftc_tag)
        self.ftc_model = AutoModelForSeq2SeqLM.from_pretrained(ftc_tag, torch_dtype=torch.float32)

        ftc_prefix = "transfer Formal to Casual: "
        src_sen = input_phrase
        sen = ftc_prefix + input_phrase
        input_ids = self.ftc_tokenizer.encode(sen, return_tensors='pt')

        self.ftc_model = self.ftc_model.to(device)
        input_ids = input_ids.to(device)

        preds = self.ftc_model.generate(
            input_ids,
            do_sample=True,
            max_length=128,
            top_k=50,
            top_p=0.95,
            early_stopping=True,
            num_return_sequences=candidates
        )

        gen_sen = set()
        for pred in preds:
            gen_sen.add(self.ftc_tokenizer.decode(pred, skip_special_tokens=True).strip())

        self.adequacy = Adequacy()
        phrases = self.adequacy.score(src_sen, list(gen_sen), filter, device)
        ranking = sorted(phrases.items(), key=lambda x: x[1], reverse=True)
        if len(ranking) > 0:
            return ranking[0][0]
        else:
            return None

    def _casual_to_formal(self, input_phrase, device, filter, candidates):
        self.ctf_tokenizer = AutoTokenizer.from_pretrained(ctf_tag)
        self.ctf_model = AutoModelForSeq2SeqLM.from_pretrained(ctf_tag, torch_dtype=torch.float32)

        ctf_prefix = "transfer Casual to Formal: "
        src_sen = input_phrase
        input_phrase = ctf_prefix + input_phrase
        input_ids = self.ctf_tokenizer.encode(input_phrase, return_tensors='pt')

        self.ctf_model = self.ctf_model.to(device)
        input_ids = input_ids.to(device)

        preds = self.ctf_model.generate(
            input_ids,
            do_sample=True,
            max_length=128,
            top_k=50,
            top_p=0.95,
            early_stopping=True,
            num_return_sequences=candidates
        )

        gen_sen = set()
        for pred in preds:
            gen_sen.add(self.ctf_tokenizer.decode(pred, skip_special_tokens=True).strip())

        self.adequacy = Adequacy()
        phrases = self.adequacy.score(src_sen, list(gen_sen), filter, device)
        ranking = sorted(phrases.items(), key=lambda x: x[1], reverse=True)
        if len(ranking) > 0:
            return ranking[0][0]
        else:
            return None

    def _active_to_passive(self, input_phrase, device):
        self.atp_tokenizer = AutoTokenizer.from_pretrained(atp_tag)
        self.atp_model = AutoModelForSeq2SeqLM.from_pretrained(atp_tag, torch_dtype=torch.float32)

        atp_prefix = "transfer Active to Passive: "
        input_phrase = atp_prefix + input_phrase
        input_ids = self.atp_tokenizer.encode(input_phrase, return_tensors='pt')

        self.atp_model = self.atp_model.to(device)
        input_ids = input_ids.to(device)

        preds = self.atp_model.generate(
            input_ids,
            do_sample=True,
            max_length=128,
            top_k=50,
            top_p=0.95,
            early_stopping=True,
            num_return_sequences=1
        )

        return self.atp_tokenizer.decode(preds[0], skip_special_tokens=True).strip()

    def _passive_to_active(self, input_phrase, device):
        self.pta_tokenizer = AutoTokenizer.from_pretrained(pta_tag)
        self.pta_model = AutoModelForSeq2SeqLM.from_pretrained(pta_tag, torch_dtype=torch.float32)

        pta_prefix = "transfer Passive to Active: "
        input_phrase = pta_prefix + input_phrase
        input_ids = self.pta_tokenizer.encode(input_phrase, return_tensors='pt')

        self.pta_model = self.pta_model.to(device)
        input_ids = input_ids.to(device)

        preds = self.pta_model.generate(
            input_ids,
            do_sample=True,
            max_length=128,
            top_k=50,
            top_p=0.95,
            early_stopping=True,
            num_return_sequences=1
        )

        return self.pta_tokenizer.decode(preds[0], skip_special_tokens=True).strip()



# coding=utf-8

import re
from os import path
from codecs import open
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the version string
with open(path.join(here, 'promptcraft/__init__.py'), encoding='utf-8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

setup(
    name='promptcraft',
    version=version,
    description='PromptCraft: A Prompt Perturbation Toolkit for Prompt Robustness Analysis',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/SuperBruceJia/promptcraft',
    author='Shuyue Jia',
    author_email='shuyuej@ieee.org',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='Prompt Perturbation',
    package_data={"":["*.py"]},
    packages=["promptcraft"],
    include_package_data=True,
    install_requires=['torch',
                      'bitsandbytes',
                      'transformers',
                      'pandas',
                      'Levenshtein',
                      'scipy',
                      'sklearn_pandas',
                      'scikit-learn',
                      'sentence_transformers',
                      'googletrans==3.1.0a0',
                      'nltk',
                      'sacremoses',
                      'sentencepiece',
                      ]
)



# coding=utf-8
#
# Copyright 2021: https://github.com/PrithivirajDamodaran/Parrot_Paraphraser
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modified by Shuyue Jia

import re
import difflib

import torch
import pandas as pd
import Levenshtein
from scipy import spatial
from scipy.special import softmax
from sklearn_pandas import DataFrameMapper
from sklearn.preprocessing import MinMaxScaler
from transformers import (
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
)
from sentence_transformers import SentenceTransformer

paraphraser_tag = "prithivida/parrot_paraphraser_on_T5"
adequacy_tag = 'prithivida/parrot_adequacy_model'
fluency_tag = 'prithivida/parrot_fluency_model'
diversity_tag = 'paraphrase-distilroberta-base-v2'


class Adequacy:

    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(adequacy_tag)
        self.model = AutoModelForSequenceClassification.from_pretrained(adequacy_tag, torch_dtype=torch.float32)

    def filter(self, input_phrase, paraphrases, threshold, device="cpu"):
        top_phrases = []
        for phrase in paraphrases:
            x = self.tokenizer(input_phrase, phrase, return_tensors='pt', max_length=128, truncation=True)
            self.model = self.model.to(device)

            logits = self.model(**x).logits
            probs = logits.softmax(dim=1)
            prob_label = probs[:, 1]

            score = prob_label.item()
            if score >= threshold:
                top_phrases.append(phrase)

        return top_phrases

    def score(self, input_phrase, paraphrases, threshold, device="cpu"):
        scores = {}
        for phrase in paraphrases:
            x = self.tokenizer(input_phrase, phrase, return_tensors='pt', max_length=128, truncation=True)
            x = x.to(device)
            self.model = self.model.to(device)

            logits = self.model(**x).logits
            probs = logits.softmax(dim=1)
            prob_label = probs[:, 1]

            score = prob_label.item()
            if score >= threshold:
                scores[phrase] = score

        return scores


class Fluency:

    def __init__(self):
        self.model = AutoModelForSequenceClassification.from_pretrained(fluency_tag, num_labels=2,
                                                                        torch_dtype=torch.float32)
        self.tokenizer = AutoTokenizer.from_pretrained(fluency_tag)

    def filter(self, paraphrases, threshold, device="cpu"):
        self.model = self.model.to(device)

        top_phrases = []
        for phrase in paraphrases:
            input_ids = self.tokenizer("Sentence: " + phrase, return_tensors='pt', truncation=True)
            input_ids = input_ids.to(device)
            prediction = self.model(**input_ids)

            scores = prediction[0][0].detach().cpu().numpy()
            scores = softmax(scores)
            score = scores[1]  # LABEL_0 = Bad Fluency, LABEL_1 = Good Fluency

            if score >= threshold:
                top_phrases.append(phrase)

        return top_phrases

    def score(self, paraphrases, threshold, device="cpu"):
        self.model = self.model.to(device)

        fluency_scores = {}
        for phrase in paraphrases:
            input_ids = self.tokenizer("Sentence: " + phrase, return_tensors='pt', truncation=True)
            input_ids = input_ids.to(device)
            prediction = self.model(**input_ids)

            scores = prediction[0][0].detach().cpu().numpy()
            scores = softmax(scores)
            score = scores[1]  # LABEL_0 = Bad Fluency, LABEL_1 = Good Fluency

            if score >= threshold:
                fluency_scores[phrase] = score

        return fluency_scores


class Diversity:

    def __init__(self):
        self.model = SentenceTransformer(diversity_tag)

    def rank(self, input_phrase, paraphrases, ranker='levenshtein'):
        if ranker == "levenshtein":
            return self.levenshtein_ranker(input_phrase, paraphrases)
        elif ranker == "euclidean":
            return self.euclidean_ranker(input_phrase, paraphrases)
        elif ranker == "diff":
            return self.diff_ranker(input_phrase, paraphrases)

    def euclidean_ranker(self, input_phrase, paraphrases):
        scores = {}
        outputs = []
        input_enc = self.model.encode(input_phrase.lower())

        for phrase in paraphrases:
            enc = self.model.encode(phrase.lower())
            l2_dis = (spatial.distance.euclidean(input_enc, enc))
            outputs.append((phrase, l2_dis))

        df = pd.DataFrame(outputs, columns=['paraphrase', 'scores'])
        fields = []
        for col in df.columns:
            if col == "scores":
                tup = ([col], MinMaxScaler())
            else:
                tup = ([col], None)

            fields.append(tup)

        mapper = DataFrameMapper(fields, df_out=True)
        for index, row in mapper.fit_transform(df.copy()).iterrows():
            scores[row['paraphrase']] = row['scores']

        return scores

    def levenshtein_ranker(self, input_phrase, paraphrases):
        scores = {}
        for phrase in paraphrases:
            distance = Levenshtein.distance(input_phrase.lower(), phrase)
            scores[phrase] = distance

        return scores

    def diff_ranker(self, input_phrase, paraphrases):
        differ = difflib.Differ()

        scores = {}
        for phrase in paraphrases:
            diff = differ.compare(input_phrase.split(), phrase.split())
            count = 0
            for d in diff:
                if "+" in d or "-" in d:
                    count += 1

            scores[phrase] = count

        return scores


class Parrot:

    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(paraphraser_tag)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(paraphraser_tag, torch_dtype=torch.float32)

        self.adequacy_score = Adequacy()
        self.fluency_score = Fluency()
        self.diversity_score = Diversity()

    def augment(
            self,
            input_phrase,
            ranker="levenshtein",
            diversity_penalty=False,
            max_return_phrases=10,
            max_length=128,
            adequacy_threshold=0.90,
            fluency_threshold=0.90
    ):
        device = "cpu"
        self.model = self.model.to(device)

        save_phrase = input_phrase
        if len(input_phrase) >= max_length:
            max_length += 32

        input_phrase = re.sub('[^a-zA-Z0-9 \?\'\-\/\:\.]', '', input_phrase)
        input_phrase = "paraphrase: " + input_phrase

        input_ids = self.tokenizer.encode(input_phrase, return_tensors='pt')
        input_ids = input_ids.to(device)

        if diversity_penalty:
            for n in range(2, 9):
                if max_return_phrases % n == 0:
                    break

            preds = self.model.generate(
                input_ids,
                do_sample=False,
                max_length=max_length,
                num_beams=max_return_phrases,
                num_beam_groups=n,
                diversity_penalty=2.0,
                early_stopping=True,
                num_return_sequences=max_return_phrases
            )

        else:
            preds = self.model.generate(
                input_ids,
                do_sample=True,
                max_length=max_length,
                top_k=50,
                top_p=0.95,
                early_stopping=True,
                num_return_sequences=max_return_phrases
            )

        paraphrases = set()

        for pred in preds:
            gen_pp = self.tokenizer.decode(pred, skip_special_tokens=True).lower()
            gen_pp = re.sub('[^a-zA-Z0-9 \?\'\-]', '', gen_pp)
            paraphrases.add(gen_pp)

        adequacy_filter = self.adequacy_score.filter(input_phrase, paraphrases, adequacy_threshold, device)
        if len(adequacy_filter) > 0:
            fluency_filter = self.fluency_score.filter(adequacy_filter, fluency_threshold, device)

            if len(fluency_filter) > 0:
                scored_phrases = self.diversity_score.rank(input_phrase, fluency_filter, ranker)

                phrases = []
                for phrase, diversity_score in scored_phrases.items():
                    phrases.append((phrase, diversity_score))

                phrases.sort(key=lambda x: x[1], reverse=True)

                return phrases

            else:
                return [(save_phrase, 0)]
